==> ppmbutton-2.1.0.tar.gz: ppmbutton-2.1.0/tglyph.c <==
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "ttf.h"

/************************************************************************/

#define DEBUG_MISC 1
#define DEBUG_FLAGS 2
#define DEBUG_POINTS 4

#ifndef DEBUG
#define DEBUG 0
#endif

/***********************************************************************
  Read a glyph from a TrueType font file
*/

int ttfCode2Glyph(
		     ttfHandle * handle,
		     int code,
		     int platformID,
		     int specificID)
{
    int n;
    int i;

    n = handle->nCharMaps;

    for (i = 0; i < n; i++) {
	ttfCharToIndexMap *ctim = handle->charToIndexMap[i];

	if (ctim->platformID == platformID &&
	    ctim->specificID == specificID) {
	    switch (ctim->format) {
		int j;

	    case 4:
		for (j = 0; j < ctim->segCount; j++) {
		    ttfFormat4Seg *f4s = ctim->format4Segs + j;
		    if (f4s->startCount <= code &&
			code <= f4s->endCount) {
			if (f4s->idRangeOffset) {
			    return f4s->glyphTable[code - f4s->startCount];
			} else {
			    return (CARD16) (code + f4s->idDelta);
			}
		    }
		}
		return -1;
	    default:
		assert(0);	/* NYI */
	    }
	}
    }
    /* Lets assume that all fonts contain a useful table for now
   * (This must be changed if this function is ever used in
   * a program which lets the user choose the code set!)
   */
    assert(0);
}

ttfGlyph *ttfReadGlyph(ttfHandle * handle, int iGlyph)
{
    ttfGlyph *glyph;
    CARD8 *p;
    CARD8 flags[1024];
    int i;
    short x, y;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "Reading glyph %d\n", iGlyph);
#endif

    if (iGlyph < 0 || iGlyph >= handle->nGlyphs) {
	fprintf(stderr, "Invalid glyph number %d\n", iGlyph);
	return NULL;
    }
    p = (char *) handle->glyphData[iGlyph];

    if ((glyph = malloc(sizeof(ttfGlyph))) == NULL) {
	perror("malloc glyph");
	return NULL;
    }
    memset(glyph, 0, sizeof(ttfGlyph));

    if (iGlyph + 1 < handle->nGlyphs &&
	handle->glyphData[iGlyph] == handle->glyphData[iGlyph + 1]) {
	/* glyph starts at same location as next one, so it must be
     * empty. This seems always to be the case for space -- maybe
     * programs know that a space is empty?
     */
	glyph->nContours = 0;
    } else {
	glyph->nContours = TTF_INT16(p);
	p += sizeof(INT16);
    }

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  No of contours: %d\n", glyph->nContours);
#endif

    if (glyph->nContours == 0) {
	fprintf(stderr, "No contours in glyph\n");
	return glyph;
    }
    glyph->xMin = TTF_FWord(p);
    p += sizeof(FWord);
    glyph->yMin = TTF_FWord(p);
    p += sizeof(FWord);
    glyph->xMax = TTF_FWord(p);
    p += sizeof(FWord);
    glyph->yMax = TTF_FWord(p);
    p += sizeof(FWord);

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  Extents: (%ld, %ld) (%ld, %ld)\n",
	    glyph->xMin, glyph->yMin,
	    glyph->xMax, glyph->yMax);
#endif

    if (glyph->nContours < 0) {
	int flags;
	int sub_glyph_index;
	int arg1, arg2;
	int xscale = 0x4000, scale01 = 0, scale10 = 0, yscale = 0x4000;
	ttfGlyph *subGlyph;
	int xoff, yoff;
	int off;

	glyph->nContours = 0;
	glyph->nPoints = 0;

	fprintf(stderr, "Composite glyph:\n");
	do {
	    flags = TTF_CARD16(p);
	    p += 2;
	    fprintf(stderr, "  flags=%x\n", flags);
	    sub_glyph_index = TTF_CARD16(p);
	    p += 2;
	    fprintf(stderr, "  sub_glyph_index=%x\n", sub_glyph_index);
	    if (flags & ARG_1_AND_2_ARE_WORDS) {
		arg1 = TTF_CARD16(p);
		p += 2;
		arg2 = TTF_CARD16(p);
		p += 2;
	    } else {
		arg1 = TTF_CARD8(p);
		p += 1;
		arg2 = TTF_CARD8(p);
		p += 1;
	    }
	    if (flags & WE_HAVE_A_SCALE) {
		xscale = yscale = TTF_INT16(p);
		p++;
	    } else if (flags & WE_HAVE_AN_X_AND_Y_SCALE) {
		xscale = TTF_INT16(p);
		p++;
		yscale = TTF_INT16(p);
		p++;
	    } else if (flags & WE_HAVE_A_TWO_BY_TWO) {
		xscale = TTF_INT16(p);
		p++;
		scale01 = TTF_INT16(p);
		p++;
		scale10 = TTF_INT16(p);
		p++;
		yscale = TTF_INT16(p);
		p++;
	    }
	    fprintf(stderr, "  scale=((%d, %d)(%d, %d))\n", xscale, scale01, scale10, yscale);

	    assert(flags & ARGS_ARE_XY_VALUES);
	    xoff = arg1;
	    yoff = arg2;

	    subGlyph = ttfReadGlyph(handle, sub_glyph_index);

	    if (flags & USE_MY_METRICS) {
		fprintf(stderr, "  USE_MY_METRICS\n");
		glyph->metricIndex = sub_glyph_index;
	    }
	    off = glyph->nContours;
	    glyph->nContours += subGlyph->nContours;
	    for (i = 0; i < subGlyph->nContours; i++) {
		glyph->endPoints[off + i] = subGlyph->endPoints[i] + glyph->nPoints;
	    }
	    off = glyph->nPoints;
	    glyph->nPoints += subGlyph->nPoints;
	    for (i = 0; i < subGlyph->nPoints; i++) {
		glyph->points[off + i].flags = subGlyph->points[i].flags;
		glyph->points[off + i].x = (xscale * subGlyph->points[i].x +
				scale01 * subGlyph->points[i].y) / 0x4000
		    + xoff;
		glyph->points[off + i].y = (scale10 * subGlyph->points[i].x +
				 yscale * subGlyph->points[i].y) / 0x4000
		    + yoff;
	    }
	} while (flags & MORE_COMPONENTS);

	return glyph;
    }
#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  Endpoints: ");
#endif
    for (i = 0; i < glyph->nContours; i++) {
	glyph->endPoints[i] = TTF_INT16(p);
	p += sizeof(INT16);

#if DEBUG & DEBUG_MISC
	fprintf(stderr, "%d ", glyph->endPoints[i]);
#endif
    }
#if DEBUG & DEBUG_MISC
    fprintf(stderr, "\n");
#endif

    glyph->nPoints = glyph->endPoints[i - 1] + 1;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  No of coordinates: %d\n", glyph->nPoints);
#endif

    i = TTF_CARD16(p);		/* instruction length */
    p += sizeof(CARD16);

#if INFO_TABLE & DEBUG_MISC
    fprintf(stderr, "  No of instructions: %d\n", i);
#endif

    p += i;			/* skip instructions */

#if DEBUG & DEBUG_FLAGS
    fprintf(stderr, "  Flags:\n");
#endif
    for (i = 0; i < glyph->nPoints; i++) {
	int flag = *p++;
#if DEBUG & DEBUG_FLAGS
	fprintf(stderr, "    %d: %x\n", i, flag);
#endif
	flags[i] = flag;
	if (flag & 8) {
	    int n = *p++;
#if DEBUG & DEBUG_FLAGS
	    fprintf(stderr, "      repeated %d times\n", n);
#endif
	    while (n--)
		flags[++i] = flag;
	}
    }
#if DEBUG & DEBUG_FLAGS
    fprintf(stderr, "\n");
#endif

    x = 0;
    for (i = 0; i < glyph->nPoints; i++) {
	short dx;

	if (flags[i] & 1)
	    glyph->points[i].flags |= TTFPOINT_ON;

	if (flags[i] & 2) {	/* x-short */
	    if (flags[i] & 16)	/* x-dual */
		dx = *p++;
	    else
		dx = -*p++;
	} else {
	    if (flags[i] & 16)	/* x-dual */
		dx = 0;
	    else {
		dx = TTF_INT16(p);
		p += sizeof(INT16);
	    }
	}

	x += dx;
	glyph->points[i].x = x;
    }

    y = 0;
    for (i = 0; i < glyph->nPoints; i++) {
	short dy;

	if (flags[i] & 4) {	/* y-short */
	    if (flags[i] & 32)	/* y-dual */
		dy = *p++;
	    else
		dy = -*p++;
	} else {
	    if (flags[i] & 32)	/* y-dual */
		dy = 0;
	    else {
		dy = TTF_INT16(p);
		p += sizeof(INT16);
	    }
	}

	y += dy;
	glyph->points[i].y = y;
    }

#if DEBUG & DEBUG_POINTS
    fprintf(stderr, "  Points:\n");
    for (i = 0; i < glyph->nPoints; i++)
	fprintf(stderr, "    %d: (%d, %d)\n", i,
		glyph->points[i].x, glyph->points[i].y);
#endif

#if DEBUG
    fprintf(stderr, "\n");
#endif

    return glyph;
}

==> ppmbutton-2.1.4.tar.gz: ppmbutton-2.1.4/tglyph.c <==
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ant/globals.h>

#include "ttf.h"

/************************************************************************/

#define DEBUG_MISC 1
#define DEBUG_FLAGS 2
#define DEBUG_POINTS 4
#define DEBUG_COMPOSITE 8

#ifndef DEBUG
#define DEBUG 0
#endif

/***********************************************************************
  Read a glyph from a TrueType font file
*/

int ttfCode2Glyph(
		     ttfHandle * handle,
		     int code,
		     int platformID,
		     int specificID)
{
    int n;
    int i;

    n = handle->nCharMaps;

    for (i = 0; i < n; i++) {
	ttfCharToIndexMap *ctim = handle->charToIndexMap[i];

	if (ctim->platformID == platformID &&
	    ctim->specificID == specificID) {
	    switch (ctim->format) {
		int j;

	    case 4:
		for (j = 0; j < ctim->segCount; j++) {
		    ttfFormat4Seg *f4s = ctim->format4Segs + j;
		    if (f4s->startCount <= code &&
			code <= f4s->endCount) {
			if (f4s->idRangeOffset) {
			    return f4s->glyphTable[code - f4s->startCount];
			} else {
			    return (CARD16) (code + f4s->idDelta);
			}
		    }
		}
		return -1;
	    default:
		assert(0);	/* NYI */
	    }
	}
    }
    fprintf(stderr, "%s: Selected ID %d:%d not found\n",
	    cmnd, platformID, specificID);
    fprintf(stderr, "\tIDs supported by this font:");
    for (i = 0; i < n; i++) {
	ttfCharToIndexMap *ctim = handle->charToIndexMap[i];

	fprintf(stderr, " %d:%d",
		ctim->platformID, ctim->specificID);
    }
    fprintf(stderr, "\n");
    exit(1);
}

ttfGlyph *ttfReadGlyph(ttfHandle * handle, int iGlyph)
{
    ttfGlyph *glyph;
    CARD8 *p;
    CARD8 flags[1024];
    int i;
    short x, y;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "Reading glyph %d\n", iGlyph);
#endif

    if (iGlyph < 0 || iGlyph >= handle->nGlyphs) {
	fprintf(stderr, "Invalid glyph number %d\n", iGlyph);
	return NULL;
    }
    p = (CARD8 *) handle->glyphData[iGlyph];

    if ((glyph = malloc(sizeof(ttfGlyph))) == NULL) {
	perror("malloc glyph");
	return NULL;
    }
    memset(glyph, 0, sizeof(ttfGlyph));

    if (iGlyph + 1 < handle->nGlyphs &&
	handle->glyphData[iGlyph] == handle->glyphData[iGlyph + 1]) {
	/* glyph starts at same location as next one, so it must be
     * empty. This seems always to be the case for space -- maybe
     * programs know that a space is empty?
     */
	glyph->nContours = 0;
    } else {
	glyph->nContours = TTF_INT16(p);
	p += sizeof(INT16);
    }

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  No of contours: %d\n", glyph->nContours);
#endif

    if (glyph->nContours == 0) {
	fprintf(stderr, "No contours in glyph\n");
	return glyph;
    }
    glyph->xMin = TTF_FWord(p);
    p += sizeof(FWord);
    glyph->yMin = TTF_FWord(p);
    p += sizeof(FWord);
    glyph->xMax = TTF_FWord(p);
    p += sizeof(FWord);
    glyph->yMax = TTF_FWord(p);
    p += sizeof(FWord);

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  Extents: (%ld, %ld) (%ld, %ld)\n",
	    glyph->xMin, glyph->yMin,
	    glyph->xMax, glyph->yMax);
#endif

    if (glyph->nContours < 0) {
	int flags;
	int sub_glyph_index;
	int arg1, arg2;
	int xscale = 0x4000, scale01 = 0, scale10 = 0, yscale = 0x4000;
	ttfGlyph *subGlyph;
	int xoff, yoff;
	int off;

	glyph->nContours = 0;
	glyph->nPoints = 0;

#if DEBUG & DEBUG_COMPOSITE
	fprintf(stderr, "Composite glyph:\n");
#endif
	do {
	    flags = TTF_CARD16(p);
	    p += 2;
#if DEBUG & DEBUG_COMPOSITE
	    fprintf(stderr, "  flags=%x\n", flags);
#endif
	    sub_glyph_index = TTF_CARD16(p);
	    p += 2;
#if DEBUG & DEBUG_COMPOSITE
	    fprintf(stderr, "  sub_glyph_index=%x\n", sub_glyph_index);
#endif
	    if (flags & ARG_1_AND_2_ARE_WORDS) {
		arg1 = TTF_INT16(p);
		p += 2;
		arg2 = TTF_INT16(p);
		p += 2;
	    } else {
		arg1 = TTF_INT8(p);
		p += 1;
		arg2 = TTF_INT8(p);
		p += 1;
	    }
	    if (flags & WE_HAVE_A_SCALE) {
		xscale = yscale = TTF_INT16(p);
		p++;
	    } else if (flags & WE_HAVE_AN_X_AND_Y_SCALE) {
		xscale = TTF_INT16(p);
		p++;
		yscale = TTF_INT16(p);
		p++;
	    } else if (flags & WE_HAVE_A_TWO_BY_TWO) {
		xscale = TTF_INT16(p);
		p++;
		scale01 = TTF_INT16(p);
		p++;
		scale10 = TTF_INT16(p);
		p++;
		yscale = TTF_INT16(p);
		p++;
	    }
#if DEBUG & DEBUG_COMPOSITE
	    fprintf(stderr, "  scale=((%d, %d)(%d, %d))\n", xscale, scale01, scale10, yscale);
#endif

	    assert(flags & ARGS_ARE_XY_VALUES);
	    xoff = arg1;
	    yoff = arg2;

	    subGlyph = ttfReadGlyph(handle, sub_glyph_index);

	    if (flags & USE_MY_METRICS) {
#if DEBUG & DEBUG_COMPOSITE
		fprintf(stderr, "  USE_MY_METRICS\n");
#endif
		glyph->metricIndex = sub_glyph_index;
	    }
	    off = glyph->nContours;
	    glyph->nContours += subGlyph->nContours;
	    for (i = 0; i < subGlyph->nContours; i++) {
		glyph->endPoints[off + i] = subGlyph->endPoints[i] + glyph->nPoints;
	    }
	    off = glyph->nPoints;
	    glyph->nPoints += subGlyph->nPoints;
	    for (i = 0; i < subGlyph->nPoints; i++) {
		glyph->points[off + i].flags = subGlyph->points[i].flags;
		glyph->points[off + i].x = (xscale * subGlyph->points[i].x +
				scale01 * subGlyph->points[i].y) / 0x4000
		    + xoff;
		glyph->points[off + i].y = (scale10 * subGlyph->points[i].x +
				 yscale * subGlyph->points[i].y) / 0x4000
		    + yoff;
	    }
	} while (flags & MORE_COMPONENTS);

	return glyph;
    }
#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  Endpoints: ");
#endif
    for (i = 0; i < glyph->nContours; i++) {
	glyph->endPoints[i] = TTF_INT16(p);
	p += sizeof(INT16);

#if DEBUG & DEBUG_MISC
	fprintf(stderr, "%d ", glyph->endPoints[i]);
#endif
    }
#if DEBUG & DEBUG_MISC
    fprintf(stderr, "\n");
#endif

    glyph->nPoints = glyph->endPoints[i - 1] + 1;

#if DEBUG & DEBUG_MISC
    fprintf(stderr, "  No of coordinates: %d\n", glyph->nPoints);
#endif

    i = TTF_CARD16(p);		/* instruction length */
    p += sizeof(CARD16);

#if INFO_TABLE & DEBUG_MISC
    fprintf(stderr, "  No of instructions: %d\n", i);
#endif

    p += i;			/* skip instructions */

#if DEBUG & DEBUG_FLAGS
    fprintf(stderr, "  Flags:\n");
#endif
    for (i = 0; i < glyph->nPoints; i++) {
	int flag = *p++;
#if DEBUG & DEBUG_FLAGS
	fprintf(stderr, "    %d: %x\n", i, flag);
#endif
	flags[i] = flag;
	if (flag & 8) {
	    int n = *p++;
#if DEBUG & DEBUG_FLAGS
	    fprintf(stderr, "      repeated %d times\n", n);
#endif
	    while (n--)
		flags[++i] = flag;
	}
    }
#if DEBUG & DEBUG_FLAGS
    fprintf(stderr, "\n");
#endif

    x = 0;
    for (i = 0; i < glyph->nPoints; i++) {
	short dx;

	if (flags[i] & 1)
	    glyph->points[i].flags |= TTFPOINT_ON;

	if (flags[i] & 2) {	/* x-short */
	    if (flags[i] & 16)	/* x-dual */
		dx = *p++;
	    else
		dx = -*p++;
	} else {
	    if (flags[i] & 16)	/* x-dual */
		dx = 0;
	    else {
		dx = TTF_INT16(p);
		p += sizeof(INT16);
	    }
	}

	x += dx;
	glyph->points[i].x = x;
    }

    y = 0;
    for (i = 0; i < glyph->nPoints; i++) {
	short dy;

	if (flags[i] & 4) {	/* y-short */
	    if (flags[i] & 32)	/* y-dual */
		dy = *p++;
	    else
		dy = -*p++;
	} else {
	    if (flags[i] & 32)	/* y-dual */
		dy = 0;
	    else {
		dy = TTF_INT16(p);
		p += sizeof(INT16);
	    }
	}

	y += dy;
	glyph->points[i].y = y;
    }

#if DEBUG & DEBUG_POINTS
    fprintf(stderr, "  Points:\n");
    for (i = 0; i < glyph->nPoints; i++)
	fprintf(stderr, "    %d: (%d, %d)\n", i,
		glyph->points[i].x, glyph->points[i].y);
#endif

#if DEBUG
    fprintf(stderr, "\n");
#endif

    return glyph;
}
